/*
 * Copyright 2009-2012 Yi-Qing WANG 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * @file EPLE_lib.cpp
 * @brief library for inpainting algorithm
 * @brief mainly written with Eigen though still in need of Newmat
 *
 * @author Yi-Qing WANG <yiqing.wang@polytechnique.edu>
 */


#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdio>

#ifndef UTIL_H
#define UTIL_H
#include "util.h"
#endif

#ifdef _OPENMP
#include <omp.h>                // OpenMP
#endif

#ifndef DATA_H
#define DATA_H
#include "DataProvider.h"       // initial mixture
#endif

#ifndef SHORT_NEWMAT
#define SHORT_NEWMAT
#include "newmat10/newmatap.h"  // NEWMAT 
typedef NEWMAT::Matrix NMatrix;
typedef NEWMAT::SymmetricMatrix NSym;
typedef NEWMAT::DiagonalMatrix NDiag;
typedef NEWMAT::ColumnVector NColV;
typedef NEWMAT::RowVector NRowV;
#endif

#include "EPLE_lib.h"
 
using namespace Eigen;
using namespace std;

//return num of factors given model and standard setting
//textural model has the same number of factors as other directional models
int ret_num_factors(
              	int model
,		PLEParams const & params
){
        //remove static if there is model culling
        static int flat_model = params.num_models - 1;
        static int textural_model = params.num_models - 2;
        if( model == flat_model )
                return 1;
        else
                return params.std_num_factors;
}

//calc model parameters given (patch, mask, model)
void calc_single_coef_posterior(
			VectorXd const & patch
,			VectorXd const & mask
,			PLEParams const & params
,			MatrixXd * all_factors
,			int model			
,			MatrixXd & second_order
,			double & cov_determinant
,			bool calc_determinant
){
	int num_factors = ret_num_factors(model, params);
	static int data_size = params.data_size;
	MatrixXd MF = mask.asDiagonal() * all_factors[model].block(0, 0, data_size, num_factors);
	double squared_sigma = pow(params.sigma, 2);
	MatrixXd inv_cov = (MF.transpose() * MF)/squared_sigma + MatrixXd::Identity(num_factors, num_factors);
	//calc the inverse of the gaussian determinant
	if( calc_determinant )
		cov_determinant = inv_cov.determinant();
	//next, look at the mean and calc the second order statistics
	VectorXd mean(num_factors+1);
	mean(num_factors) = 1;
	mean.head(num_factors) = inv_cov.ldlt().solve(MF.transpose()*(patch - all_factors[model].col(num_factors)))/squared_sigma;
	second_order.block(0, 0, num_factors, num_factors) = inv_cov.inverse();
	second_order += mean * mean.transpose();	
}


//calc patch responsibility and update mixing weight, as well as sigma 
double infer_responsibility(
			MatrixXd * all_factors
,			VectorXd ** patches
,			VectorXd ** masks
,			PLEParams const & params
,			VectorXd & mixing_weight
,			VectorXd **& responsibility
,			MatrixXd **& second_order	
){
		int map_rows = params.map_rows;
		int map_cols = params.map_cols;
		//allocate memory 
		if( responsibility == NULL ){
			responsibility = new VectorXd * [map_rows];
			second_order = new MatrixXd * [map_rows];
			for( int row = 0; row < map_rows; row++ ){
				responsibility[row] = new VectorXd [map_cols];
				second_order[row] = new MatrixXd [map_cols];
			}
		}
		//estimate responsibility, sigma and mixing weights	
		double squared_sigma = pow(params.sigma, 2);
		double estimated_var = 0;
		double total_available = 0;
		int num_models = params.num_models;
		int textural_model = num_models - 2;
		int flat_model = num_models - 1;
		//with no information available, I default to textural model
		VectorXd default_resp;
		default_resp.setZero(num_models);
		default_resp(flat_model) = 1;
		//only store second order statistics for the textural model for it's the most expensive to calculate
		#pragma omp parallel for schedule(static) reduction(+:total_available,estimated_var)
		for( int row = 0; row < map_rows; row++ )	
			for( int col = 0; col < map_cols; col++ ){
				if(masks[row][col].sum() == 0){
					responsibility[row][col] = default_resp;
					continue;
				}	
				//log responsibility and sample variance
				VectorXd log_resp, sample_var;
				sample_var.setConstant(num_models, patches[row][col].squaredNorm());
				//-500 is just one arbitrary number
				log_resp.setConstant(num_models, -500);
				//number of observable pixels
				total_available += masks[row][col].sum();
				bool inf_flag[num_models];
				//I'll leave out the common factors in the log responsibility calculation
				for( int model = 0; model < num_models; model++ ){
					//additional step to prevent exception
					inf_flag[model] = false;
					//testing
	   	                      //  if( isnan( log(mixing_weight(model)) ) ){
	   	                      //          //sanity check
	   	                      //          if( mixing_weight(model) < 0 )
	   	                      //                  cout << "ERROR: log prior is Nan because the prior is negative, please check." << endl;
	   	                      //          else
	   	                      //                  cout << "INFO: model " << model << " has a prior equal to absolute 0." << endl;
	   	                      //          inf_flag[model] = true;
	   	                      //          continue;
	   	                      //  }
			
					//calculate the second order statistics	
					int num_factors = ret_num_factors(model, params);
					double determinant;
					MatrixXd second_order_stat;
					second_order_stat.setZero(num_factors+1, num_factors+1);
					calc_single_coef_posterior(patches[row][col], masks[row][col], params, all_factors, model, second_order_stat, determinant);
					if( model == textural_model )
						second_order[row][col] = second_order_stat;

					//first portion of log responsibility
					log_resp(model) = log(mixing_weight(model)) - 0.5*(log(determinant) + second_order_stat.trace());
					MatrixXd MF = masks[row][col].asDiagonal() * all_factors[model];
					//because of both matrices being symmetric, I can do this
					double trace = second_order_stat.cwiseProduct(MF.transpose()*MF).sum();
					log_resp(model) -= 0.5*trace/squared_sigma;
					sample_var(model) += trace;
					//cross
					double cross = patches[row][col].dot(MF*second_order_stat.col(num_factors));
					log_resp(model) += cross/squared_sigma;
					sample_var(model) -= 2*cross;
				}
				//try to use the full numerical range
				VectorXd max_setter;
				max_setter.setConstant(num_models, 700 - log_resp.maxCoeff());
				log_resp += max_setter;
				for( int model = 0; model < num_models; model++ )
					log_resp(model) = inf_flag[model] ? 0 : exp(log_resp(model));
				//normalize
				responsibility[row][col] = log_resp/log_resp.sum();
				estimated_var += sample_var.dot(responsibility[row][col]);
			}
			//update mixing weights
			mixing_weight.setZero(num_models);
			for( int row = 0; row < map_rows; row++ )	
				for( int col = 0; col < map_cols; col++ )
					mixing_weight += responsibility[row][col];
			mixing_weight /= map_rows * map_cols;
			cout << "INFO: updated mixing_weight : " << endl;
			cout << fixed << setprecision(2) << mixing_weight.transpose() << endl;
			//return the sigma, rather than update it, gives more freedom at testing
			return sqrt(max(1e-40, estimated_var)/total_available);
}

//update factors
void calc_model_factors(
			VectorXd ** patches
,			VectorXd ** masks
,			PLEParams & params
,			VectorXd & mixing_weight
,			VectorXd ** responsibility	
,			MatrixXd ** second_order
,			MatrixXd *& all_factors
){
	int map_rows = params.map_rows;
	int map_cols = params.map_cols;
	int num_models = params.num_models;
	int textural_model = num_models - 2;
	int data_size = params.data_size;
	#pragma omp parallel for schedule(static)
	for( int model = 0; model < num_models; model++ ){
		if( mixing_weight[model] < 1e-20 ){
			cout << "INFO: model " << model << " not accounting for data, we don't update its factors." << endl;
			continue;
		}
                //we'll solve for each row a linear equation
		int num_factors = ret_num_factors(model, params);
		//right hand side of the equation
		MatrixXd RHS;
		RHS.setZero(data_size, num_factors+1);
		//left hand side of the equation
		MatrixXd LHS[data_size];
                for( int row_id = 0; row_id < data_size; row_id++ )
                        LHS[row_id].setZero(num_factors+1, num_factors+1);
		VectorXd counter = VectorXd::Zero(data_size);
   		//build the equation
		for( int row = 0; row < map_rows; row++ )
			for( int col = 0; col < map_cols; col++ ){
				//fully masked patch doesn't contribute here
				if(masks[row][col].sum() == 0)
					continue;
				MatrixXd tmpL;
				if( model != textural_model ){
					MatrixXd second_order_stat;
					double determinant;
					second_order_stat.setZero(num_factors+1, num_factors+1);
					calc_single_coef_posterior(patches[row][col], masks[row][col], params, all_factors, model, second_order_stat, determinant, false);
                 	        	tmpL = second_order_stat*responsibility[row][col][model];
				}else
					tmpL = second_order[row][col]*responsibility[row][col][model];
                 	        RHS += patches[row][col]*tmpL.row(num_factors);
				counter += masks[row][col];
                 	        for( int row_id = 0; row_id < data_size; row_id++ ){
                 	                if( masks[row][col][row_id] > 0 )
                 	                        LHS[row_id] += tmpL;
                 	        }
                }
		//under the i.i.d. assumption, a valid estimate should be returned at each row
                for( int row = 0; row < data_size; row++ )
                        if( counter[row] > 0 )
                                all_factors[model].row(row) = (LHS[row].ldlt().solve(RHS.row(row).transpose())).transpose();
                        else
                                cout << "ERROR:: not enough data for factor updating for model " << model << " at row " << row << endl;
	}
}

//EM algorithm which inpaints too
void EM_Learn( 
		VectorXd **& patches
,		VectorXd **& masks
,		int num_iterations
,		int ch
,		int num_channels
,		PLEParams & params
,		MatrixXd *& all_factors
,		VectorXd & mixing_weight
,		VectorXd **& responsibility
,		MatrixXi & patch_map
,		bool demo_mode
,		MatrixXd *& intermediate_recovery	
){
	//update sigma or keep it as it is at the beginning
	bool update_sigma = true;
	cout << "INFO: EM routine starts" << endl;
	if(demo_mode){
		cout << "INFO: I'm in the demo mode" << endl;
		//thus only inpaint once
		intermediate_recovery = new MatrixXd [1];
	}else{
		cout << "INFO: I'm in the testing mode" << endl;
		intermediate_recovery = new MatrixXd [num_iterations];
	}
	bool learn = num_channels == 1 ? true : (ch == 1) || !demo_mode;
	MatrixXd ** second_order = NULL;
	//keep a copy for at inpaint_patches, I actually modify patches as well as masks
	VectorXd ** patch_copies = NULL;
	VectorXd ** mask_copies = NULL;
	if(!demo_mode){
		patch_copies = new VectorXd * [params.map_rows];
		mask_copies = new VectorXd * [params.map_rows];
		for( int row = 0; row < params.map_rows; row++ ){
			mask_copies[row] = new VectorXd [params.map_cols];
			patch_copies[row] = new VectorXd [params.map_cols];
			for( int col = 0; col < params.map_cols; col++ ){
				mask_copies[row][col] = masks[row][col];
				patch_copies[row][col] = patches[row][col];
			}
		}
	}
	for( int iter = 0; iter < num_iterations; iter++ ){
		if(learn){
			times("");
			double sigma = infer_responsibility(all_factors, patches, masks, params, mixing_weight, responsibility, second_order);
			times("RESP: ");
			if( update_sigma ){
				cout << "INFO: sigma updated from " << params.sigma << " to " << sigma << endl;
				params.sigma = sigma;
			}
		}
		if(!demo_mode || iter == num_iterations-1){
			//display the patch map
			produce_patch_map(responsibility, params, patch_map);
			VectorXd ** restored_patches = NULL;	
			inpaint_patches(patches, masks, responsibility, all_factors, patch_map, params, restored_patches);
			if(!demo_mode){
				//in demo mode, since these are no longer used, I don't care for them.
				for( int row = 0; row < params.map_rows; row++ )
					for( int col = 0; col < params.map_cols; col++ ){
						patches[row][col] = patch_copies[row][col];
						masks[row][col] = mask_copies[row][col];
					}
			}
			int image_rows = params.image_rows;
			int image_cols = params.image_cols;
			int patch_size = sqrt(params.data_size);
			int overlap = patch_size - 1;
			MatrixXd recovered(image_rows, image_cols);
			patches2image(restored_patches, overlap, patch_size, recovered);
			if(!demo_mode)
				intermediate_recovery[iter] = recovered;
			else
				intermediate_recovery[0] = recovered;
			for( int row = 0; row < params.map_rows; row++ )
				delete [] restored_patches[row];
			delete [] restored_patches;
		}
		//only update if necessary
		if( learn && iter < num_iterations - 1 ){
			times("");
			calc_model_factors(patches, masks, params, mixing_weight, responsibility, second_order, all_factors);
			times("FACTORs: ");
		}
	}
	//release
	if(learn){
		for( int row = 0; row < params.map_rows; row++ )
			delete [] second_order[row];
		delete [] second_order;
	}
	if(!demo_mode){
		for( int row = 0; row < params.map_rows; row++ ){
			delete [] patch_copies[row];
			delete [] mask_copies[row];
		}	
		delete [] patch_copies;
		delete [] mask_copies;
	}
}

//produce and display the patch map
void produce_patch_map(
		VectorXd ** responsibility
,		PLEParams const & params
,		MatrixXi & patch_map
){
	int map_rows = params.map_rows;
	int map_cols = params.map_cols;
	int num_models = params.num_models;
	#pragma omp parallel for schedule(static)
	for( int row = 0; row < map_rows; row++ )
		for( int col = 0; col < map_cols; col++ ){
			double highest;
			int best_model;
			for( int model = 0; model < num_models; model++ ){
				double resp = responsibility[row][col][model];
				if(model == 0 || highest < resp){
					highest = resp;
					best_model = model;
				}
			}
			patch_map(row, col) = best_model;
		}
	//print
	print_patch_map(patch_map);
}

//fill in the holes with Wiener filtering
void inpaint_patches(    
			VectorXd **& patches
,                       VectorXd **& masks
,                       VectorXd ** responsibility
,                       MatrixXd * all_factors
,                       MatrixXi & patch_map
,			PLEParams & params
,                       VectorXd **& restored_patches
){
	int map_rows = params.map_rows;
	int map_cols = params.map_cols;
	//allocate memory
	restored_patches = new VectorXd * [map_rows];
	for( int row = 0; row < map_rows; row++ )
		restored_patches[row] = new VectorXd [map_cols];
	//parameters
	int data_size = params.data_size;
	int patch_size = sqrt(data_size);
	VectorXd ones;
	ones.setOnes(data_size);
	//an arbitrarily fixed parameter
	double data_fit_control = 1;
	bool speedup = true;
	int num_models = params.num_models;
	int textural_model = num_models - 2;
	//start filtering
	bool stop = false;
	bool first_round = true;
	while(!stop){
		int unpainted = 0;
		#pragma omp parallel for schedule(static) reduction(+:unpainted)
		for( int row = 0; row < map_rows; row++ )
			for( int col = 0; col < map_cols; col++ ){
				//don't handle a patch either because it's fully visible or fully masked
				if(masks[row][col].sum() == 0){
					unpainted += 1;
					restored_patches[row][col] = patches[row][col];
					continue;
				}
				if(masks[row][col].sum() == data_size){
					restored_patches[row][col] = patches[row][col];
					continue;
				}
				VectorXd Wiener_restored;
				Wiener_restored.setZero(data_size);
				int start_model, end_model;
				if( first_round ){
					start_model = speedup ? patch_map(row, col) : 0;
					end_model = speedup ? patch_map(row, col)+1 : num_models;
				}else{
					//really a hack 
					start_model = textural_model;
					end_model = textural_model + 1;
				}
				double normalizer = 0;
				for( int model = start_model; model < end_model; model++ ){
					//ignoring this won't have any numerical impact
					if( responsibility[row][col][model] < 1e-4 )
						continue;
					normalizer += responsibility[row][col][model];
					int num_factors = ret_num_factors(model, params);
					MatrixXd F = all_factors[model].block(0, 0, data_size, num_factors);
					MatrixXd MF = masks[row][col].asDiagonal()*F;
					MatrixXd kernel = data_fit_control*MF.transpose()*MF + MatrixXd::Identity(num_factors, num_factors);
					VectorXd mu = all_factors[model].col(num_factors);
					Wiener_restored += responsibility[row][col][model]*(mu+data_fit_control*F*kernel.ldlt().solve(MF.transpose()*(patches[row][col]-mu))); 
				}
				Wiener_restored /= normalizer;
				//retrieve the estimated pixels
				restored_patches[row][col] = (ones-masks[row][col]).asDiagonal()*Wiener_restored + patches[row][col];
				//update masks
				masks[row][col] = ones;
			}
		cout << "INFO: " << unpainted << " patches remain to be inpainted." << endl;
		stop = unpainted == 0 ? true : false;
		if( !stop ){
			//if there are patches being totally masked at the beginning
			static int image_rows = params.image_rows;
			static int image_cols = params.image_cols;
			MatrixXd restored(image_rows, image_cols);
			static int overlap = patch_size - 1;
			patches2image(restored_patches, overlap, patch_size, restored, false);
			MatrixXd new_mask(image_rows, image_cols);
			patches2image(masks, overlap, patch_size, new_mask, false);
			//partially restore the image
			for( int row = 0; row < image_rows; row++ )
				for( int col = 0; col < image_cols; col++ ){
					restored(row, col) = new_mask(row, col) > 0 ? restored(row, col)/new_mask(row, col) : 0;
					new_mask(row, col) = new_mask(row, col) > 0 ? 1 : 0;
				}
			//renew patches and masks
			for( int row  = 0; row < map_rows; row++ ){
				delete [] masks[row];
				delete [] patches[row];
			}
			delete [] masks;
			delete [] patches;
			masks = image2patches(new_mask, image_rows, image_cols, overlap, patch_size);
			patches = image2patches(restored, image_rows, image_cols, overlap, patch_size);
		}
	}
}

//deduce factors and their eigen-values from a covariance matrix
//this routine is needed as several places in this implementation
void deduce_eigens(
                MatrixXd & cov_mat
,		PLEParams const & params
,               int model
,               MatrixXd & eig_vecs
,               VectorXd & eig_vals
){
		int data_size = params.data_size;
                //standard SVD: Methods provided in Eigen can be flaky
                //Revert to what I know is reliable: NEWMAT
                NMatrix cov_container(data_size, data_size);
                cov_container << cov_mat.data();
                NSym sym_cov(data_size);
                sym_cov << cov_container;
                NMatrix U(data_size, data_size);
                NDiag D(data_size);
                SVD(sym_cov, D, U);
                //convert to the other world
                //the index in NEWMAT library starts at 1 instead of 0
                for( int i = 0; i < data_size; i++ ){
                        eig_vals(i) = D(i+1);
                        for( int j = 0; j < data_size; j++ )
                                eig_vecs(i, j) = U(i+1,j+1);
                }
}

//initial mixture
void read_config(
		PLEParams const & params
,               VectorXd *& prior
,               MatrixXd *& all_factors
){
        bool regularize_cov = true;
        if( regularize_cov )
                cout << "INFO: covariance regularization is planned." << endl;
        else
                cout << "INFO: covariance regularization is NOT planned." << endl;
	int data_size = params.data_size;
        int patch_size = (int)sqrt(data_size);
	int num_models = params.num_models;
        //initial GMM setup
        DataProvider my_data;
        //mixing weights
	prior = new VectorXd [1];
        prior[0] = Map<VectorXd>(my_data.GetPrior(patch_size), num_models, 1);
        //model factor loading + model mean
        all_factors = new MatrixXd [num_models];
        for( int model = 0; model < num_models; model++ ){
                //determine the number of factors for this model
                int num_factors = ret_num_factors(model, params);
		//one additional column for the model mean
		MatrixXd factors(data_size, num_factors+1);
		//set the model mean
                factors.col(num_factors) = Map<MatrixXd>(my_data.GetMu(patch_size, model), data_size, 1);
		//calc now the rest, which uses svd
                MatrixXd cov_mat = Map<MatrixXd>(my_data.GetCov( patch_size, model ), data_size, data_size);
		MatrixXd eig_vecs(data_size, data_size);
                VectorXd eig_vals(data_size);
                deduce_eigens(cov_mat, params, model, eig_vecs, eig_vals);
                //regularize the covariance matrices
                double average_noise_energy = 0;
                int redundancy = data_size - num_factors;
                if( regularize_cov && redundancy > 0 )
                        average_noise_energy = eig_vals.tail( redundancy ).sum()/(double)redundancy;
                VectorXd cst;
                cst.setConstant(num_factors, average_noise_energy);
                VectorXd sqrt_eig_vals = (eig_vals.head( num_factors ) - cst).cwiseSqrt();
		factors.block(0, 0, data_size, num_factors) = eig_vecs.block(0, 0, data_size, num_factors) * sqrt_eig_vals.asDiagonal();
		all_factors[model] = factors;
        }
        //for testing
        if( false ){
                cout << "INFO: here comes the initial prior" << endl;
                cout << fixed << setprecision(2) << prior[0] << endl;
                cout << "INFO: and they sum to " << prior[0].sum() << endl;
        }
}



void inpaint_image_sub_routine(
			MatrixXd * & image
,			int num_channels
,			MatrixXd & mask
,			PLEParams & params
,			int num_iterations
,			bool demo_mode
,			MatrixXd ** & restored_image	
){
	//read config
	int num_models = params.num_models;
	//get patch masks
	int image_rows = mask.rows();
	int image_cols = mask.cols();
	int data_size = params.data_size;
	int patch_size = sqrt(data_size);
	int overlap = patch_size - 1;
	//start inpainting
	VectorXd ** responsibility = NULL;
	MatrixXi patch_map(params.map_rows, params.map_cols);
	restored_image = new MatrixXd * [num_channels];
	int map_rows = params.map_rows;
	int map_cols = params.map_cols;
	MatrixXd * all_factors = NULL;
	VectorXd * mixing_weight = NULL;
	//adjust channel priority 
	int my_channels[3] = {1, 2, 0};
	for( int index = 0; index < num_channels; index++ ){
		int ch = num_channels == 1 ? 0 : my_channels[index];
		//read config only if necessary
		if(index == 0 || !demo_mode){
			if( all_factors != NULL ){
				delete [] all_factors;
				delete [] mixing_weight;
			}
			read_config(params, mixing_weight, all_factors);
			params.sigma = 1;
		}
		VectorXd ** patches = image2patches(image[ch], image_rows, image_cols, overlap, patch_size);
		VectorXd ** masks = image2patches(mask, image_rows, image_cols, overlap, patch_size);
		cout << "INFO: CH = " << ch << endl;
		EM_Learn(patches, masks, num_iterations, ch, num_channels, params, all_factors, mixing_weight[0], responsibility, patch_map, demo_mode, restored_image[ch]);
		//release
		for( int row = 0; row < map_rows; row++ ){
			delete [] patches[row];
			delete [] masks[row];
		}
		delete [] patches;
		delete [] masks;
		//release
		if(!demo_mode || index == num_channels - 1){
			for( int row = 0; row < map_rows; row++ )
				delete [] responsibility[row];
			delete [] responsibility;
			responsibility = NULL;
			delete [] all_factors;
			all_factors = NULL;
			delete [] mixing_weight;
		}
	}
}


//main routine
int inpaint_image_main_routine( 
			const char * input
,			const char * mask_name
,			const char * output
,			int num_iterations
,			int data_size			
,			bool demo_mode	
){
	//read in the masked image
  	int num_channels, image_rows, image_cols;
        MatrixXd * full = NULL;
        imread(input, image_rows, image_cols, num_channels, full);
        if( full == NULL )
                fail("Unable to get the image");
			
	//read in the mask
	MatrixXd * mask = NULL;
	int mask_channels;
	imread(mask_name, image_rows, image_cols, mask_channels, mask);
	if( mask == NULL )
		fail("Unable to get the mask");
	if( mask[0].sum() == 255*image_cols*image_rows )
		fail("Nothing visible left to inpaint from");

	//transform the mask to conform to my convention 
	MatrixXd opposite;
	opposite.setConstant(image_rows, image_cols, 255);
	mask[0] -= opposite;
	mask[0] /= -255;

	//mask the image and output an example
	MatrixXd * masked_image = new MatrixXd [num_channels];
	for( int ch = 0; ch < num_channels; ch++ ){
		masked_image[ch] = full[ch].cwiseProduct(mask[0]);
		full[ch] = masked_image[ch];
	}
	imwrite("masked.png", full, num_channels);

	//build the param struct
	PLEParams params;
	params.sigma = 1;
	params.num_models = 20;
	params.data_size = data_size;
	params.std_num_factors = data_size/2;
	int patch_size = sqrt(data_size);
	int overlap = patch_size - 1;
	params.map_rows = num_patches(image_rows, patch_size, overlap);
	params.map_cols = num_patches(image_cols, patch_size, overlap);
	params.image_rows = image_rows;
	params.image_cols = image_cols;

	//inpaint
	MatrixXd ** restored_image = NULL;
	inpaint_image_sub_routine(masked_image, num_channels, mask[0], params, num_iterations, demo_mode, restored_image);

	//write out more images in the testing mode
	if(!demo_mode && num_iterations > 1){
		int num_words = (int)strlen(output);
	        char image_name[num_words-3];
	        strncpy(image_name, output, num_words-4);
	        image_name[num_words-4] = '\0';
	        char buffer[100];
		for( int iter = 0; iter < num_iterations - 1; iter++ ){
	       		sprintf(buffer, "%s_%02d.png", image_name, iter);
	       	 	const char * fname = buffer;
			MatrixXd * container = new MatrixXd [num_channels];
			for( int ch = 0; ch < num_channels; ch++ )
				container[ch] = restored_image[ch][iter];
	       	 	imwrite(fname, container, num_channels);
		}
	}
	//write out the final one
	int label = demo_mode ? 0 : num_iterations - 1;
	MatrixXd * container = new MatrixXd [num_channels];
	for( int ch = 0; ch < num_channels; ch++ ){
		container[ch] = restored_image[ch][label];
		delete [] restored_image[ch];
	}
	delete [] restored_image;
	imwrite(output, container, num_channels);
	
	return EXIT_SUCCESS;
}
