/*
 * Copyright 2009-2012 Yi-Qing WANG 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file EPLE_lib.h
 * @brief header file for EPLE denoising algorithm
 * @author Yi-Qing WANG <yiqing.wang@polytechnique.edu>
 */

#ifndef EIGEN_H
#define EIGEN_H
#include <Eigen/Dense>          // Eigen
#endif

using namespace Eigen;

/**
 * @brief struct contains the parameters for the algorithm 
 *
 * @param sigma estimated image noise level
 * @param num_models the number of mixture components 
 * @param std_num_factors number of factors in an oriented model 
 * @param data_size the number of pixels in a patch 
 * @param map_rows the number of rows in the patch map
 * @param map_cols the number of columns in the patch map
 * @param image_rows the number of pixels in each image column
 * @param image_cols the number of pixels in each image row 
 */
struct PLEParams{
	double sigma;
	int num_models;
	int std_num_factors;
	int data_size;
	int map_rows;
	int map_cols;
	int image_rows;
	int image_cols;
};

/** @brief return the number of factors given a model */
int ret_num_factors(
              	int model
,		PLEParams const & params
);

/**
 * @brief calculate the posterior gaussian parameter given an observed patch 
 *
 * @param patch the current patch 
 * @param mask the mask applied to the current patch 
 * @param params the parameters of the algorithm setup and the image 
 * @param all_factors factors of all components in the mixture 
 * @param model the current model 
 * @param second_order the covariance matrix of (factor coefficient, model mean)
 * @param cov_determinant the inverse determinant of the coefficient covariance
 * @param calc_determinant whether to calculate the determinant 
 */
void calc_single_coef_posterior(
			VectorXd const & patch
,			VectorXd const & mask
,			PLEParams const & params
,			MatrixXd * all_factors
,			int model			
,			MatrixXd & second_order
,			double & cov_determinant
,			bool calc_determinant = true
);

/**
 * @brief infer the posterior model distribution given an observation 
 *
 * @param all_factors factors of all components in the mixture 
 * @param patches all patches extracted from the image to inpaint 
 * @param masks the masks associated with the extracted patches 
 * @param params the parameters of the algorithm setup and the image
 * @param mixing_weight the mixture model prior 
 * @param responsibility posterior model distributions of all the observations 
 * @param second_order the covariance matrix of (factor coefficient, model mean)
 */
double infer_responsibility(
			MatrixXd * all_factors
,			VectorXd ** patches
,			VectorXd ** masks
,			PLEParams const & params
,			VectorXd & mixing_weight
,			VectorXd **& responsibility
,			MatrixXd **& second_order	
);

/**
 * @brief update model factors for all the mixture components 
 *
 * @param patches all patches extracted from the image to inpaint 
 * @param masks the masks associated with the extracted patches 
 * @param params the parameters of the algorithm setup and the image
 * @param mixing_weight the mixture model prior 
 * @param responsibility posterior model distributions of all the observations 
 * @param second_order the covariance matrix of (factor coefficient, model mean)
 * @param all_factors factors of all components in the mixture 
 */
void calc_model_factors(
			VectorXd ** patches
,			VectorXd ** masks
,			PLEParams & params
,			VectorXd & mixing_weight
,			VectorXd ** responsibility
,			MatrixXd ** second_order	
,			MatrixXd *& all_factors
);

/**
 * @brief standard factor model EM updating plus patch inpainting 
 *
 * @param patches all patches extracted from the image to inpaint 
 * @param masks the masks associated with the extracted patches 
 * @param num_iterations the number of EM iterations  
 * @param ch the channel I'm dealing with 
 * @param num_channels the number of channels allowing to tell RGB from gray images 
 * @param params the parameters of the algorithm setup and the image
 * @param all_factors factors of all components in the mixture 
 * @param mixing_weight the mixture model prior 
 * @param responsibility posterior model distributions of all the observations 
 * @param patch_map the map that associates the most appropriate model with a given patch 
 * @param demo_mode whether I'm in demo mode or testing mode 
 * @param intermediate_recovery restored images of the channel being looked at 
 */
void EM_Learn( 
		VectorXd **& patches
,		VectorXd **& masks
,		int num_iterations
,		int ch
,		int num_channels
,		PLEParams & params
,		MatrixXd *& all_factors
,		VectorXd & mixing_weight
,		VectorXd **& responsibility
,		MatrixXi & patch_map	
,		bool demo_mode
,		MatrixXd *& intermediate_recovery	
);

/**
 * @brief create the patch map according to responsibility and print it out 
 *
 * @param responsibility posterior model distributions of all the observations 
 * @param params the parameters of the algorithm setup and the image
 * @param patch_map the map that associates the most appropriate model with a given patch 
 */
void produce_patch_map(
		VectorXd ** responsibility
,		PLEParams const & params
,		MatrixXi & patch_map
);

/**
 * @brief a routine that does the actual inpainting with Wiener filtering 
 *
 * @param patches all patches extracted from the image to inpaint 
 * @param masks the masks associated with the extracted patches 
 * @param responsibility posterior model distributions of all the observations 
 * @param all_factors factors of all components in the mixture 
 * @param patch_map the map that associates the most appropriate model with a given patch 
 * @param params the parameters of the algorithm setup and the image
 * @param restored_patches inpainted patches 
 */
void inpaint_patches(    
			VectorXd **& patches
,                       VectorXd **& masks
,                       VectorXd ** responsibility
,                       MatrixXd * all_factors
,                       MatrixXi & patch_map
,			PLEParams & params
,                       VectorXd **& restored_patches
);

/**
 * @brief svd 
 *
 * @param cov_mat a covariance matrix 
 * @param params the parameters of the algorithm setup and the image
 * @param model the current model 
 * @param eig_vecs the eigenvectors of the covariance matrix 
 * @param eig_vals the eigenvalues of the covariance matrix 
 */
void deduce_eigens(
                MatrixXd & cov_mat
,		PLEParams const & params
,               int model
,               MatrixXd & eig_vecs
,               VectorXd & eig_vals
);

/**
 * @brief read in the initial mixture configuration 
 *
 * @param params the parameters of the algorithm setup and the image
 * @param prior mixing weights 
 * @param all_factors factors of all components in the mixture 
 */
void read_config(
		PLEParams const & params
,               VectorXd *& prior
,               MatrixXd *& all_factors
);

/**
 * @brief read in the initial mixture configuration 
 *
 * @param the image to inpaint 
 * @param num_channels RGB or gray 
 * @param mask the mask expressed as a black and white image 
 * @param params the parameters of the algorithm setup and the image
 * @param num_iterations the number of EM iterations 
 * @param demo_mode in demo or testing mode 
 * @param restored_image the restored image 
 */
void inpaint_image_sub_routine(
			MatrixXd * & image
,			int num_channels
,			MatrixXd & mask
,			PLEParams & params
,			int num_iterations
,			bool demo_mode
,			MatrixXd ** & restored_image	
);

/**
 * @brief read in the initial mixture configuration 
 *
 * @param input the input image name
 * @param mask the input mask name 
 * @param output the output inpainted image name 
 * @param num_iterations the number of EM iterations 
 * @param data_size the number of pixels in a patch 
 * @param demo_mode in demo or testing mode 
 */
int inpaint_image_main_routine( 
			const char * input
,			const char * mask
,			const char * output
,			int num_iterations
,			int data_size			
,			bool demo_mode
);
